@@ -581,7 +581,7 @@ int QQmlPrivate::qmlregister(RegistrationType type, void *data)
         auto revisions = prepareRevisions(type.instanceMetaObject, added) + furtherRevisions;
         uniqueRevisions(&revisions, type.version, added);
 
-        for (QTypeRevision revision : qAsConst(revisions)) {
+        for (QTypeRevision revision : std::as_const(revisions)) {
             if (revision.hasMajorVersion() && revision.majorVersion() > type.version.majorVersion())
                 break;
 
@@ -624,7 +624,7 @@ int QQmlPrivate::qmlregister(RegistrationType type, void *data)
         revisions.append(added);
         uniqueRevisions(&revisions, type.version, added);
 
-        for (QTypeRevision revision : qAsConst(revisions)) {
+        for (QTypeRevision revision : std::as_const(revisions)) {
             if (revision < added)
                 continue;
             if (revision.hasMajorVersion() && revision.majorVersion() > type.version.majorVersion())
@@ -976,7 +976,8 @@ static ObjectLookupResult initObjectLookup(
         // & 1 to tell the gc that this is not heap allocated; see markObjects in qv4lookup_p.h
         l->qobjectFallbackLookup.metaObject = quintptr(metaObject) + 1;
         l->qobjectFallbackLookup.coreIndex = coreIndex;
-        l->qobjectFallbackLookup.notifyIndex = property.notifySignalIndex();
+        l->qobjectFallbackLookup.notifyIndex =
+                QMetaObjectPrivate::signalIndex(property.notifySignal());
         l->qobjectFallbackLookup.isConstant = property.isConstant() ? 1 : 0;
         return ObjectLookupResult::Fallback;
     }
@@ -997,7 +998,7 @@ static bool initValueLookup(QV4::Lookup *l, QV4::ExecutableCompilationUnit *comp
     const QByteArray name = compilationUnit->runtimeStrings[l->nameIndex]->toQString().toUtf8();
     const int coreIndex = metaObject->indexOfProperty(name.constData());
     QMetaType lookupType = metaObject->property(coreIndex).metaType();
-    if (type.isValid() && lookupType != type)
+    if (!isTypeCompatible(type, lookupType))
         return false;
     l->qgadgetLookup.metaObject = quintptr(metaObject) + 1;
     l->qgadgetLookup.coreIndex = coreIndex;
