@@ -38,6 +38,10 @@ QQmlJSTypeResolver::QQmlJSTypeResolver(QQmlJSImporter *importer)
     m_varType = builtinTypes[u"QVariant"_s].scope;
     m_jsValueType = builtinTypes[u"QJSValue"_s].scope;
 
+    QQmlJSScope::Ptr emptyType = QQmlJSScope::create();
+    emptyType->setAccessSemantics(QQmlJSScope::AccessSemantics::None);
+    m_emptyType = emptyType;
+
     QQmlJSScope::Ptr emptyListType = QQmlJSScope::create();
     emptyListType->setInternalName(u"void*"_s);
     emptyListType->setAccessSemantics(QQmlJSScope::AccessSemantics::Sequence);
@@ -70,7 +74,7 @@ QQmlJSTypeResolver::QQmlJSTypeResolver(QQmlJSImporter *importer)
 
     m_jsGlobalObject = importer->jsGlobalObject();
     auto numberMethods = m_jsGlobalObject->methods(u"Number"_s);
-    Q_ASSERT(numberMethods.length() == 1);
+    Q_ASSERT(numberMethods.size() == 1);
     m_numberPrototype = numberMethods[0].returnType()->baseType();
     Q_ASSERT(m_numberPrototype);
     Q_ASSERT(m_numberPrototype->internalName() == u"NumberPrototype"_s);
@@ -169,7 +173,10 @@ QQmlJSScope::ConstPtr QQmlJSTypeResolver::typeForConst(QV4::ReturnedValue rv) co
         return realType();
 
     if (value.isNull())
-        return jsPrimitiveType();
+        return nullType();
+
+    if (value.isEmpty())
+        return emptyType();
 
     return {};
 }
@@ -753,6 +760,7 @@ QQmlJSScope::ConstPtr QQmlJSTypeResolver::genericType(const QQmlJSScope::ConstPt
             return type;
         if (const QQmlJSScope::ConstPtr valueType = type->valueType())
             return listType(genericType(valueType), UseQObjectList);
+        return m_variantListType;
     }
 
     return m_varType;
@@ -1013,6 +1021,10 @@ QQmlJSRegisterContent QQmlJSTypeResolver::memberType(const QQmlJSScope::ConstPtr
 {
     QQmlJSRegisterContent result;
 
+    // If we got a plain type reference we have to check the enums of the _scope_.
+    if (equals(type, metaObjectType()))
+        return {};
+
     if (equals(type, jsValueType())) {
         QQmlJSMetaProperty prop;
         prop.setPropertyName(name);
