@@ -1,41 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the QtQuick module of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:LGPL$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU Lesser General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU Lesser
-** General Public License version 3 as published by the Free Software
-** Foundation and appearing in the file LICENSE.LGPL3 included in the
-** packaging of this file. Please review the following information to
-** ensure the GNU Lesser General Public License version 3 requirements
-** will be met: https://www.gnu.org/licenses/lgpl-3.0.html.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 2.0 or (at your option) the GNU General
-** Public license version 3 or any later version approved by the KDE Free
-** Qt Foundation. The licenses are as published by the Free Software
-** Foundation and appearing in the file LICENSE.GPL2 and LICENSE.GPL3
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-2.0.html and
-** https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR LGPL-3.0-only OR GPL-2.0-only OR GPL-3.0-only
 
 #include <private/qquickshadereffect_p.h>
 #include <private/qsgcontextplugin_p.h>
@@ -100,14 +64,21 @@ QT_BEGIN_NAMESPACE
     The following inputs are predefined:
 
     \list
-    \li vec4 qt_Vertex - vertex position, the top-left vertex has
+    \li vec4 qt_Vertex with location 0 - vertex position, the top-left vertex has
        position (0, 0), the bottom-right (\l{Item::width}{width},
        \l{Item::height}{height}).
-    \li vec2 qt_MultiTexCoord0 - texture coordinate, the top-left
+    \li vec2 qt_MultiTexCoord0 with location 1 - texture coordinate, the top-left
        coordinate is (0, 0), the bottom-right (1, 1). If \l supportsAtlasTextures
        is true, coordinates will be based on position in the atlas instead.
     \endlist
 
+    \note It is only the vertex input location that matters in practice. The
+    names are freely changeable, while the location must always be \c 0 for
+    vertex position, \c 1 for texture coordinates. However, be aware that this
+    applies to vertex inputs only, and is not necessarily true for output
+    variables from the vertex shader that are then used as inputs in the
+    fragment shader (typically, the interpolated texture coordinates).
+
     The following uniforms are predefined:
 
     \list
@@ -123,6 +94,9 @@ QT_BEGIN_NAMESPACE
 
     \note The uniform block layout qualifier must always be \c std140.
 
+    \note Unlike vertex inputs, the predefined names (qt_Matrix, qt_Opacity)
+    must not be changed.
+
     In addition, any property that can be mapped to a GLSL type can be made
     available to the shaders. The following list shows how properties are
     mapped:
@@ -215,7 +189,6 @@ QT_BEGIN_NAMESPACE
             mat4 qt_Matrix;
             float qt_Opacity;
         };
-        out gl_PerVertex { vec4 gl_Position; };
         void main() {
             coord = qt_MultiTexCoord0;
             gl_Position = qt_Matrix * qt_Vertex;
@@ -261,6 +234,18 @@ QT_BEGIN_NAMESPACE
     or the opacity, because at run time there is one single uniform buffer that
     is exposed to both the vertex and fragment shader.
 
+    \warning Unlike with vertex inputs, passing data between the vertex and
+    fragment shader may, depending on the underlying graphics API, require the
+    same names to be used, a matching location is not always sufficient. Most
+    prominently, when specifying a fragment shader while relying on the default,
+    built-in vertex shader, the texture coordinates are passed on as \c
+    qt_TexCoord0 at location \c 0, and therefore it is strongly advised that the
+    fragment shader declares the input with the same name
+    (qt_TexCoord0). Failing to do so may lead to issues on some platforms, for
+    example when running with a non-core profile OpenGL context where the
+    underlying GLSL shader source code has no location qualifiers and matching
+    is based on the variable names during to shader linking process.
+
     \section1 ShaderEffect and Item Layers
 
     The ShaderEffect type can be combined with \l {Item Layers} {layered items}.
@@ -275,7 +260,8 @@ QT_BEGIN_NAMESPACE
               id: layerRoot
               layer.enabled: true
               layer.effect: ShaderEffect {
-              fragmentShader: "effect.frag.qsb"
+                 fragmentShader: "effect.frag.qsb"
+              }
           }
           \endqml
 
@@ -354,7 +340,159 @@ QT_BEGIN_NAMESPACE
     corner. For non-linear vertex transformations, like page curl, you can
     specify a fine grid of vertices by specifying a \l mesh resolution.
 
-    \sa {Item Layers}
+    \section1 Migrating From Qt 5
+
+    For Qt 5 applications with ShaderEffect items the migration to Qt 6 involves:
+    \list
+    \li Moving the shader code to separate \c{.vert} and \c{.frag} files,
+    \li updating the shaders to Vulkan-compatible GLSL,
+    \li running the \c qsb tool on them,
+    \li including the resulting \c{.qsb} files in the executable with the Qt resource system,
+    \li and referencing the file in the \l vertexShader and \l fragmentShader properties.
+    \endlist
+
+    As described in the \l{Qt Shader Tools} module some of these steps can be
+    automated by letting CMake invoke the \c qsb tool at build time. See \l{Qt
+    Shader Tools Build System Integration} for more information and examples.
+
+    When it comes to updating the shader code, below is an overview of the
+    commonly required changes.
+
+    \table
+    \header
+    \li Vertex shader in Qt 5
+    \li Vertex shader in Qt 6
+    \row
+    \li \badcode
+        attribute highp vec4 qt_Vertex;
+        attribute highp vec2 qt_MultiTexCoord0;
+        varying highp vec2 coord;
+        uniform highp mat4 qt_Matrix;
+        void main() {
+            coord = qt_MultiTexCoord0;
+            gl_Position = qt_Matrix * qt_Vertex;
+        }
+    \endcode
+    \li \badcode
+        #version 440
+        layout(location = 0) in vec4 qt_Vertex;
+        layout(location = 1) in vec2 qt_MultiTexCoord0;
+        layout(location = 0) out vec2 coord;
+        layout(std140, binding = 0) uniform buf {
+            mat4 qt_Matrix;
+            float qt_Opacity;
+        };
+        void main() {
+            coord = qt_MultiTexCoord0;
+            gl_Position = qt_Matrix * qt_Vertex;
+        }
+    \endcode
+    \endtable
+
+    The conversion process mostly involves updating the code to be compatible
+    with
+    \l{https://github.com/KhronosGroup/GLSL/blob/master/extensions/khr/GL_KHR_vulkan_glsl.txt}{GL_KHR_vulkan_glsl}.
+    It is worth noting that Qt Quick uses a subset of the features provided by
+    GLSL and Vulkan, and therefore the conversion process for typical
+    ShaderEffect shaders is usually straightforward.
+
+    \list
+
+    \li The \c version directive should state \c 440 or \c 450, although
+    specifying other GLSL version may work too, because the
+    \l{https://github.com/KhronosGroup/GLSL/blob/master/extensions/khr/GL_KHR_vulkan_glsl.txt}{GL_KHR_vulkan_glsl}
+    extension is written for GLSL 140 and higher.
+
+    \li Inputs and outputs must use the modern GLSL \c in and \c out keywords.
+    In addition, specifying a location is required. The input and output
+    location namespaces are separate, and therefore assigning locations
+    starting from 0 for both is safe.
+
+    \li When it comes to vertex shader inputs, the only possibilities with
+    ShaderEffect are location \c 0 for vertex position (traditionally named \c
+    qt_Vertex) and location \c 1 for texture coordinates (traditionally named
+    \c qt_MultiTexCoord0).
+
+    \li The vertex shader outputs and fragment shader inputs are up to the
+    shader code to define. The fragment shader must have a \c vec4 output at
+    location 0 (typically called \c fragColor). For maximum portability, vertex
+    outputs and fragment inputs should use both the same location number and the
+    same name. When specifying only a fragment shader, the texture coordinates
+    are passed in from the built-in vertex shader as \c{vec2 qt_TexCoord0} at
+    location \c 0, as shown in the example snippets above.
+
+    \li Uniform variables outside a uniform block are not legal. Rather,
+    uniform data must be declared in a uniform block with binding point \c 0.
+
+    \li The uniform block is expected to use the std140 qualifier.
+
+    \li At run time, the vertex and fragment shader will get the same uniform
+    buffer bound to binding point 0. Therefore, as a general rule, the uniform
+    block declarations must be identical between the shaders. This also
+    includes members that are not used in one of the shaders. The member names
+    must match, because with some graphics APIs the uniform block is converted
+    to a traditional struct uniform, transparently to the application.
+
+    \li When providing one of the shaders only, watch out for the fact that the
+    built-in shaders expect \c qt_Matrix and \c qt_Opacity at the top of the
+    uniform block. (more precisely, at offset 0 and 64, respectively) As a
+    general rule, always include these as the first and second members in the
+    block.
+
+    \li In the example the uniform block specifies the block name \c buf. This
+    name can be changed freely, but must match between the shaders. Using an
+    instance name, such as \c{layout(...) uniform buf { ... } instance_name;}
+    is optional. When specified, all accesses to the members must be qualified
+    with instance_name.
+
+    \endlist
+
+    \table
+    \header
+    \li Fragment shader in Qt 5
+    \li Fragment shader in Qt 6
+    \row
+    \li \badcode
+        varying highp vec2 coord;
+        uniform lowp float qt_Opacity;
+        uniform sampler2D src;
+        void main() {
+            lowp vec4 tex = texture2D(src, coord);
+            gl_FragColor = vec4(vec3(dot(tex.rgb,
+                                vec3(0.344, 0.5, 0.156))),
+                                     tex.a) * qt_Opacity;
+        }
+    \endcode
+    \li \badcode
+        #version 440
+        layout(location = 0) in vec2 coord;
+        layout(location = 0) out vec4 fragColor;
+        layout(std140, binding = 0) uniform buf {
+            mat4 qt_Matrix;
+            float qt_Opacity;
+        };
+        layout(binding = 1) uniform sampler2D src;
+        void main() {
+            vec4 tex = texture(src, coord);
+            fragColor = vec4(vec3(dot(tex.rgb,
+                             vec3(0.344, 0.5, 0.156))),
+                                  tex.a) * qt_Opacity;
+        }
+    \endcode
+    \endtable
+
+    \list
+
+    \li Precision qualifiers (\c lowp, \c mediump, \c highp) are not currently used.
+
+    \li Calling built-in GLSL functions must follow the modern GLSL names, most
+    prominently, \c{texture()} instead of \c{texture2D()}.
+
+    \li Samplers must use binding points starting from 1.
+
+    \endlist
+
+    \sa {Item Layers}, {QSB Manual}, {Qt Shader Tools Build System Integration}
 */
 
 
@@ -991,6 +1129,8 @@ QSGNode *QQuickShaderEffectImpl::handleUpdatePaintNode(QSGNode *oldNode, QQuickI
     sd.fragment.shader = &m_shaders[Fragment];
     sd.fragment.dirtyConstants = &m_dirtyConstants[Fragment];
     sd.fragment.dirtyTextures = &m_dirtyTextures[Fragment];
+    sd.materialTypeCacheKey = m_item->window();
+
     node->syncMaterial(&sd);
 
     if (m_dirty & QSGShaderEffectNode::DirtyShaderMesh) {
@@ -1069,6 +1209,7 @@ bool QQuickShaderEffectImpl::updateUniformValue(const QByteArray &name, const QV
     sd.fragment.shader = &m_shaders[Fragment];
     sd.fragment.dirtyConstants = &dirtyConstants[Fragment];
     sd.fragment.dirtyTextures = {};
+    sd.materialTypeCacheKey = m_item->window();
 
     node->syncMaterial(&sd);
 
@@ -1080,7 +1221,7 @@ void QQuickShaderEffectImpl::handleItemChange(QQuickItem::ItemChange change, con
     // Move the window ref.
     if (change == QQuickItem::ItemSceneChange) {
         for (int shaderType = 0; shaderType < NShader; ++shaderType) {
-            for (const auto &vd : qAsConst(m_shaders[shaderType].varData)) {
+            for (const auto &vd : std::as_const(m_shaders[shaderType].varData)) {
                 if (vd.specialType == QSGShaderEffectNode::VariableData::Source) {
                     QQuickItem *source = qobject_cast<QQuickItem *>(qvariant_cast<QObject *>(vd.value));
                     if (source) {
@@ -1122,7 +1263,7 @@ void QQuickShaderEffectImpl::disconnectSignals(Shader shaderType)
         if (mapper)
             QObjectPrivate::disconnect(m_item, mapper->signalIndex(), &a);
     }
-    for (const auto &vd : qAsConst(m_shaders[shaderType].varData)) {
+    for (const auto &vd : std::as_const(m_shaders[shaderType].varData)) {
         if (vd.specialType == QSGShaderEffectNode::VariableData::Source) {
             QQuickItem *source = qobject_cast<QQuickItem *>(qvariant_cast<QObject *>(vd.value));
             if (source) {
@@ -1136,7 +1277,7 @@ void QQuickShaderEffectImpl::disconnectSignals(Shader shaderType)
 
 void QQuickShaderEffectImpl::clearMappers(QQuickShaderEffectImpl::Shader shaderType)
 {
-    for (auto *mapper : qAsConst(m_mappers[shaderType])) {
+    for (auto *mapper : std::as_const(m_mappers[shaderType])) {
         if (mapper)
             mapper->destroyIfLastRef();
     }
@@ -1155,27 +1296,13 @@ static inline QVariant getValueFromProperty(QObject *item, const QMetaObject *it
     return value;
 }
 
-struct ShaderInfoCache
+using QQuickShaderInfoCache = QHash<QUrl, QSGGuiThreadShaderEffectManager::ShaderInfo>;
+Q_GLOBAL_STATIC(QQuickShaderInfoCache, shaderInfoCache)
+
+void qtquick_shadereffect_purge_gui_thread_shader_cache()
 {
-    bool contains(const QUrl &key) const
-    {
-        return m_shaderInfoCache.contains(key);
-    }
-
-    QSGGuiThreadShaderEffectManager::ShaderInfo value(const QUrl &key) const
-    {
-        return m_shaderInfoCache.value(key);
-    }
-
-    void insert(const QUrl &key, const QSGGuiThreadShaderEffectManager::ShaderInfo &value)
-    {
-        m_shaderInfoCache.insert(key, value);
-    }
-
-    QHash<QUrl, QSGGuiThreadShaderEffectManager::ShaderInfo> m_shaderInfoCache;
-};
-
-Q_GLOBAL_STATIC(ShaderInfoCache, shaderInfoCache)
+    shaderInfoCache()->clear();
+}
 
 bool QQuickShaderEffectImpl::updateShader(Shader shaderType, const QUrl &fileUrl)
 {
@@ -1187,14 +1314,15 @@ bool QQuickShaderEffectImpl::updateShader(Shader shaderType, const QUrl &fileUrl
 
     disconnectSignals(shaderType);
 
-    m_shaders[shaderType].shaderInfo = QSGGuiThreadShaderEffectManager::ShaderInfo();
+    m_shaders[shaderType].shaderInfo.variables.clear();
     m_shaders[shaderType].varData.clear();
 
     if (!fileUrl.isEmpty()) {
         const QQmlContext *context = qmlContext(m_item);
         const QUrl loadUrl = context ? context->resolvedUrl(fileUrl) : fileUrl;
-        if (shaderInfoCache()->contains(loadUrl)) {
-            m_shaders[shaderType].shaderInfo = shaderInfoCache()->value(loadUrl);
+        auto it = shaderInfoCache()->constFind(loadUrl);
+        if (it != shaderInfoCache()->cend()) {
+            m_shaders[shaderType].shaderInfo = *it;
             m_shaders[shaderType].hasShaderCode = true;
         } else {
             // Each prepareShaderCode call needs its own work area, hence the
@@ -1223,7 +1351,7 @@ bool QQuickShaderEffectImpl::updateShader(Shader shaderType, const QUrl &fileUrl
             // provided and monitored like with an application-provided shader.
             QSGGuiThreadShaderEffectManager::ShaderInfo::Variable v;
             v.name = QByteArrayLiteral("source");
-            v.bindPoint = 0; // fake
+            v.bindPoint = 1; // fake, must match the default source bindPoint in qquickshadereffectnode.cpp
             v.type = texturesSeparate ? QSGGuiThreadShaderEffectManager::ShaderInfo::Texture
                                       : QSGGuiThreadShaderEffectManager::ShaderInfo::Sampler;
             m_shaders[shaderType].shaderInfo.variables.append(v);
@@ -1274,16 +1402,13 @@ void QQuickShaderEffectImpl::updateShaderVars(Shader shaderType)
 
     const bool texturesSeparate = mgr->hasSeparateSamplerAndTextureObjects();
 
-    const int varCount = m_shaders[shaderType].shaderInfo.variables.count();
+    const int varCount = m_shaders[shaderType].shaderInfo.variables.size();
     m_shaders[shaderType].varData.resize(varCount);
 
-    // Reuse signal mappers as much as possible since the mapping is based on
-    // the index and shader type which are both constant.
-    if (m_mappers[shaderType].count() < varCount)
-        m_mappers[shaderType].resize(varCount);
+    // Recreate signal mappers when the shader has changed.
+    clearMappers(shaderType);
 
-    auto *engine = qmlEngine(m_item);
-    QQmlPropertyCache *propCache = engine ? QQmlData::ensurePropertyCache(engine, m_item) : nullptr;
+    QQmlPropertyCache::ConstPtr propCache = QQmlData::ensurePropertyCache(m_item);
 
     if (!m_itemMetaObject)
         m_itemMetaObject = m_item->metaObject();
@@ -1323,7 +1448,7 @@ void QQuickShaderEffectImpl::updateShaderVars(Shader shaderType)
 
         // Find the property on the ShaderEffect item.
         int propIdx = -1;
-        QQmlPropertyData *pd = nullptr;
+        const QQmlPropertyData *pd = nullptr;
         if (propCache) {
             pd = propCache->property(QLatin1String(v.name), nullptr, nullptr);
             if (pd) {
@@ -1336,13 +1461,11 @@ void QQuickShaderEffectImpl::updateShaderVars(Shader shaderType)
                 if (pd->notifyIndex() == -1) {
                     qWarning("QQuickShaderEffect: property '%s' does not have notification method!", v.name.constData());
                 } else {
-                    auto *&mapper = m_mappers[shaderType][i];
-                    if (!mapper) {
-                        const int mappedId = indexToMappedId(shaderType, i);
-                        mapper = new QtPrivate::EffectSlotMapper([this, mappedId](){
-                            this->propertyChanged(mappedId);
-                        });
-                    }
+                    const int mappedId = indexToMappedId(shaderType, i);
+                    auto mapper = new QtPrivate::EffectSlotMapper([this, mappedId](){
+                        this->propertyChanged(mappedId);
+                    });
+                    m_mappers[shaderType].append(mapper);
                     mapper->setSignalIndex(m_itemMetaObject->property(propIdx).notifySignal().methodIndex());
                     Q_ASSERT(m_item->metaObject() == m_itemMetaObject);
                     bool ok = QObjectPrivate::connectImpl(m_item, pd->notifyIndex(), m_item, nullptr, mapper,
@@ -1376,7 +1499,7 @@ void QQuickShaderEffectImpl::updateShaderVars(Shader shaderType)
 bool QQuickShaderEffectImpl::sourceIsUnique(QQuickItem *source, Shader typeToSkip, int indexToSkip) const
 {
     for (int shaderType = 0; shaderType < NShader; ++shaderType) {
-        for (int idx = 0; idx < m_shaders[shaderType].varData.count(); ++idx) {
+        for (int idx = 0; idx < m_shaders[shaderType].varData.size(); ++idx) {
             if (shaderType != typeToSkip || idx != indexToSkip) {
                 const auto &vd(m_shaders[shaderType].varData[idx]);
                 if (vd.specialType == QSGShaderEffectNode::VariableData::Source && qvariant_cast<QObject *>(vd.value) == source)
@@ -1391,7 +1514,7 @@ std::optional<int> QQuickShaderEffectImpl::findMappedShaderVariableId(const QByt
 {
     for (int shaderType = 0; shaderType < NShader; ++shaderType) {
         const auto &vars = m_shaders[shaderType].shaderInfo.variables;
-        for (int idx = 0; idx < vars.count(); ++idx) {
+        for (int idx = 0; idx < vars.size(); ++idx) {
             if (vars[idx].name == name)
                 return indexToMappedId(shaderType, idx);
         }
@@ -1407,10 +1530,11 @@ void QQuickShaderEffectImpl::propertyChanged(int mappedId)
     const auto &v(m_shaders[type].shaderInfo.variables[idx]);
     auto &vd(m_shaders[type].varData[idx]);
 
+    QVariant oldValue = vd.value;
     vd.value = getValueFromProperty(m_item, m_itemMetaObject, v.name, vd.propertyIndex);
 
     if (vd.specialType == QSGShaderEffectNode::VariableData::Source) {
-        QQuickItem *source = qobject_cast<QQuickItem *>(qvariant_cast<QObject *>(vd.value));
+        QQuickItem *source = qobject_cast<QQuickItem *>(qvariant_cast<QObject *>(oldValue));
         if (source) {
             if (m_item->window())
                 QQuickItemPrivate::get(source)->derefWindow();
