@@ -2,14 +2,14 @@
 // This code is governed by the BSD license found in the LICENSE file.
 
 /*---
-es5id: 13.1-39-s
+es5id: 13.1-41-s
 description: >
     StrictMode - SyntaxError is thrown if 'arguments' occurs as the
-    function name of a FunctionDeclaration in strict eval code
+    Identifier of a FunctionExpression in strict eval code
 flags: [noStrict]
 ---*/
 
-
+        var _13_1_41_s = {};
 assert.throws(SyntaxError, function() {
-            eval("'use strict'; function arguments() { };")
+            eval("'use strict'; _13_1_41_s.x = function arguments() {};");
 });
