@@ -1,30 +1,5 @@
-/****************************************************************************
-**
-** Copyright (C) 2016 The Qt Company Ltd.
-** Contact: https://www.qt.io/licensing/
-**
-** This file is part of the test suite of the Qt Toolkit.
-**
-** $QT_BEGIN_LICENSE:GPL-EXCEPT$
-** Commercial License Usage
-** Licensees holding valid commercial Qt licenses may use this file in
-** accordance with the commercial license agreement provided with the
-** Software or, alternatively, in accordance with the terms contained in
-** a written agreement between you and The Qt Company. For licensing terms
-** and conditions see https://www.qt.io/terms-conditions. For further
-** information use the contact form at https://www.qt.io/contact-us.
-**
-** GNU General Public License Usage
-** Alternatively, this file may be used under the terms of the GNU
-** General Public License version 3 as published by the Free Software
-** Foundation with exceptions as appearing in the file LICENSE.GPL3-EXCEPT
-** included in the packaging of this file. Please review the following
-** information to ensure the GNU General Public License requirements will
-** be met: https://www.gnu.org/licenses/gpl-3.0.html.
-**
-** $QT_END_LICENSE$
-**
-****************************************************************************/
+// Copyright (C) 2016 The Qt Company Ltd.
+// SPDX-License-Identifier: LicenseRef-Qt-Commercial OR GPL-3.0-only WITH Qt-GPL-exception-1.0
 #ifndef TESTTYPES_H
 #define TESTTYPES_H
 
@@ -192,6 +167,22 @@ private:
 QML_DECLARE_TYPE(MyQmlObject)
 QML_DECLARE_TYPEINFO(MyQmlObject, QML_HAS_ATTACHED_PROPERTIES)
 
+class MyQmlObjectWithAttachedCounter : public QObject
+{
+    Q_OBJECT
+public:
+    MyQmlObjectWithAttachedCounter(QObject *parent = nullptr) : QObject(parent) { }
+    static int attachedCount;
+
+    static MyAttachedObject *qmlAttachedProperties(QObject *other)
+    {
+        ++MyQmlObjectWithAttachedCounter::attachedCount;
+        return new MyAttachedObject(other);
+    }
+};
+QML_DECLARE_TYPE(MyQmlObjectWithAttachedCounter)
+QML_DECLARE_TYPEINFO(MyQmlObjectWithAttachedCounter, QML_HAS_ATTACHED_PROPERTIES)
+
 class MyGroupedObject : public QObject
 {
     Q_OBJECT
@@ -1063,7 +1054,7 @@ class MyUncreateableBaseClass : public QObject
     Q_PROPERTY(bool prop2 READ prop2 WRITE setprop2 REVISION 1)
     Q_PROPERTY(bool prop3 READ prop3 WRITE setprop3 REVISION 1)
 public:
-    explicit MyUncreateableBaseClass(bool /* arg */, QObject *parent = 0)
+    explicit MyUncreateableBaseClass(bool /* arg */, QObject *parent = nullptr)
         : QObject(parent), _prop1(false), _prop2(false), _prop3(false)
     {
     }
@@ -1085,7 +1076,7 @@ class MyCreateableDerivedClass : public MyUncreateableBaseClass
     Q_PROPERTY(bool prop2 READ prop2 WRITE setprop2 REVISION 1)
 
 public:
-    MyCreateableDerivedClass(QObject *parent = 0)
+    MyCreateableDerivedClass(QObject *parent = nullptr)
         : MyUncreateableBaseClass(true, parent)
     {
     }
@@ -1098,7 +1089,7 @@ class MyExtendedUncreateableBaseClass : public QObject
     Q_PROPERTY(bool prop2 READ prop2 WRITE setprop2 REVISION 1)
     Q_PROPERTY(bool prop3 READ prop3 WRITE setprop3 REVISION 1)
 public:
-    explicit MyExtendedUncreateableBaseClass(QObject *parent = 0)
+    explicit MyExtendedUncreateableBaseClass(QObject *parent = nullptr)
         : QObject(parent), _prop1(false), _prop2(false), _prop3(false)
     {
     }
@@ -1119,7 +1110,7 @@ class MyExtendedUncreateableBaseClassExtension : public QObject
     Q_OBJECT
     Q_PROPERTY(bool prop4 READ prop4 WRITE setprop4)
 public:
-    explicit MyExtendedUncreateableBaseClassExtension(QObject *parent = 0)
+    explicit MyExtendedUncreateableBaseClassExtension(QObject *parent = nullptr)
         : QObject(parent), _prop4(false)
     {
     }
@@ -1135,7 +1126,7 @@ class MyExtendedCreateableDerivedClass : public MyExtendedUncreateableBaseClass
     Q_PROPERTY(bool prop5 READ prop5 WRITE setprop5)
 
 public:
-    MyExtendedCreateableDerivedClass(QObject *parent = 0)
+    MyExtendedCreateableDerivedClass(QObject *parent = nullptr)
         : MyExtendedUncreateableBaseClass(parent), _prop5(false)
     {
     }
@@ -1366,7 +1357,7 @@ class SimpleObjectExtension : public QObject
     Q_OBJECT
     Q_PROPERTY(int extendedProperty READ extendedProperty WRITE setExtendedProperty NOTIFY extendedPropertyChanged)
 public:
-    SimpleObjectExtension(QObject *parent = 0)
+    SimpleObjectExtension(QObject *parent = nullptr)
         : QObject(parent)
         , m_extendedProperty(1584)
     {}
@@ -1438,6 +1429,12 @@ private:
     QObjectList m_list;
 };
 
+class ImmediateProperties : public QObject
+{
+    Q_OBJECT
+    Q_CLASSINFO("ImmediatePropertyNames", "objectName")
+};
+
 namespace ScopedEnumsWithNameClash
 {
 Q_NAMESPACE
@@ -1461,13 +1458,57 @@ public:
     enum class OtherScopedEnum : int { ScopedVal1, ScopedVal2, ScopedVal3 };
 };
 
+class AttachedType : public QObject
+{
+    Q_OBJECT
+    QML_ANONYMOUS
+    Q_PROPERTY(
+            QString attachedName READ attachedName WRITE setAttachedName NOTIFY attachedNameChanged)
+
+    QString m_name;
+
+public:
+    AttachedType(QObject *parent = nullptr) : QObject(parent) { }
+
+    QString attachedName() const { return m_name; }
+    void setAttachedName(const QString &name)
+    {
+        if (name != m_name) {
+            m_name = name;
+            Q_EMIT attachedNameChanged();
+        }
+    }
+Q_SIGNALS:
+    void attachedNameChanged();
+};
+
 class Extension : public QObject
 {
     Q_OBJECT
-    Q_PROPERTY(int extension READ extension CONSTANT)
+    Q_PROPERTY(int extension READ extension WRITE setExtension NOTIFY extensionChangedWithValue FINAL)
+
+    QML_ATTACHED(AttachedType)
 public:
     Extension(QObject *parent = nullptr) : QObject(parent) {}
-    int extension() const { return 42; }
+    int extension() const { return ext; }
+    void setExtension(int e) {
+        if (e != ext) {
+            ext = e;
+            emit extensionChanged();
+            emit extensionChangedWithValue(e);
+        }
+    }
+    Q_INVOKABLE int invokable() { return 123; }
+
+    static AttachedType *qmlAttachedProperties(QObject *object) { return new AttachedType(object); }
+
+Q_SIGNALS:
+    void extensionChanged();
+    void extensionChangedWithValue(int value);
+public slots:
+    int slot() { return 456; }
+private:
+    int ext = 42;
 };
 
 class Extended : public QObject
@@ -1563,6 +1604,41 @@ public:
     int own() const { return 93; }
 };
 
+class ExtendedByNamespaceInParent : public ExtendedByNamespace
+{
+    Q_OBJECT
+    QML_ELEMENT
+public:
+    ExtendedByNamespaceInParent(QObject *parent = nullptr) : ExtendedByNamespace(parent) { }
+};
+
+class ExtendedNamespaceByObject : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_EXTENDED_NAMESPACE(Extension)
+
+    Q_PROPERTY(QString dummy READ dummy CONSTANT)
+    Q_PROPERTY(int extension READ extension WRITE setExtension NOTIFY extensionChanged)
+
+    int m_ext = 0;
+
+public:
+    ExtendedNamespaceByObject(QObject *parent = nullptr) : QObject(parent) {}
+    QString dummy() const { return QStringLiteral("dummy"); }
+    int extension() const { return m_ext; }
+    void setExtension(int e)
+    {
+        if (e != m_ext) {
+            m_ext = e;
+            Q_EMIT extensionChanged();
+        }
+    }
+
+Q_SIGNALS:
+    void extensionChanged();
+};
+
 class FactorySingleton : public QObject
 {
     Q_OBJECT
@@ -1670,6 +1746,14 @@ public:
     int d() const { return 22; }
 };
 
+class IndirectExtensionB : public ExtensionB
+{
+    Q_OBJECT
+    QML_ANONYMOUS
+public:
+    IndirectExtensionB(QObject *parent = nullptr) : ExtensionB(parent) { }
+};
+
 class MultiExtensionParent : public QObject
 {
     Q_OBJECT
@@ -1700,6 +1784,150 @@ public:
     int g() const { return 44; }
 };
 
+class MultiExtensionIndirect : public MultiExtensionParent
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_EXTENDED(IndirectExtensionB)
+
+    Q_PROPERTY(int b READ b CONSTANT) // won't be able to use ExtensionB, so provide own property
+
+    Q_PROPERTY(int e READ e CONSTANT)
+    Q_PROPERTY(int c READ c CONSTANT)
+    Q_PROPERTY(int g READ g CONSTANT)
+public:
+    MultiExtensionIndirect(QObject *parent = nullptr) : MultiExtensionParent(parent) { }
+
+    int b() const { return 77; }
+
+    int e() const { return 'e'; }
+    int c() const { return 'c'; }
+    int g() const { return 44; }
+};
+
+class ExtendedInParent : public MultiExtensionParent
+{
+    Q_OBJECT
+    QML_ELEMENT
+    // properties from base type: p, c, f
+    // properties from base type's extension: a, c, d, f, g
+
+    Q_PROPERTY(int c READ c CONSTANT) // be evil: overwrite base type extension's property
+public:
+    ExtendedInParent(QObject *parent = nullptr) : MultiExtensionParent(parent) { }
+
+    int c()
+    {
+        Q_UNREACHABLE();
+        return 1111;
+    }
+};
+
+class ExtendedByIndirect : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_EXTENDED(IndirectExtensionB)
+    // properties from extension's base type: b, c, d
+public:
+    ExtendedByIndirect(QObject *parent = nullptr) : QObject(parent) { }
+};
+
+class ExtendedInParentByIndirect : public ExtendedByIndirect
+{
+    Q_OBJECT
+    QML_ELEMENT
+    // properties from base type's extension's base type: b, c, d
+public:
+    ExtendedInParentByIndirect(QObject *parent = nullptr) : ExtendedByIndirect(parent) { }
+};
+
+class MultiExtensionThreeExtensions : public MultiExtension
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_EXTENDED(Extension)
+public:
+    MultiExtensionThreeExtensions(QObject *parent = nullptr) : MultiExtension(parent) { }
+};
+
+class MultiExtensionWithoutExtension : public MultiExtension
+{
+    Q_OBJECT
+    QML_ANONYMOUS
+public:
+    MultiExtensionWithoutExtension(QObject *parent = nullptr) : MultiExtension(parent) { }
+};
+
+class MultiExtensionWithExtensionInBaseBase : public MultiExtensionWithoutExtension
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_EXTENDED(Extension)
+public:
+    MultiExtensionWithExtensionInBaseBase(QObject *parent = nullptr)
+        : MultiExtensionWithoutExtension(parent)
+    {
+    }
+};
+
+class RevisionedExtension : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(int extension READ extension WRITE setExtension REVISION(1, 0))
+public:
+    RevisionedExtension(QObject *parent = nullptr) : QObject(parent) {}
+    int extension() const { return m_ext; }
+    void setExtension(int e) { m_ext = e; }
+private:
+    int m_ext = 42;
+};
+
+class ExtendedWithRevisionOld : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_EXTENDED(RevisionedExtension)
+    QML_ADDED_IN_VERSION(0, 5)
+public:
+    ExtendedWithRevisionOld(QObject *parent = nullptr) : QObject(parent) { }
+};
+
+class ExtendedWithRevisionNew : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    QML_EXTENDED(RevisionedExtension)
+    QML_ADDED_IN_VERSION(1, 0)
+public:
+    ExtendedWithRevisionNew(QObject *parent = nullptr) : QObject(parent) { }
+};
+
+class MyExtendedGroupedObject : public MyGroupedObject
+{
+    Q_OBJECT
+    QML_ANONYMOUS
+    QML_EXTENDED(Extension)
+    Q_PROPERTY(int value2 READ value2 WRITE setValue2)
+    int m_value2 = 0;
+public:
+    int value2() const { return m_value2; }
+    void setValue2(int v) { m_value2 = v; }
+};
+
+class ExtendedInGroup : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+    Q_PROPERTY(MyExtendedGroupedObject *group READ group)
+
+    MyExtendedGroupedObject m_group;
+public:
+    ExtendedInGroup(QObject *parent = nullptr) : QObject(parent) { }
+
+    MyExtendedGroupedObject *group() { return &m_group; }
+};
+
 class StringSignaler : public QObject
 {
     Q_OBJECT
@@ -1711,6 +1939,296 @@ signals:
     void signal(QJSValue value);
 };
 
+class EnumList : public QObject
+{
+    Q_OBJECT
+    QML_ELEMENT
+public:
+    enum Enum { Alpha, Beta, Gamma };
+    Q_ENUM(Enum)
+
+    Q_INVOKABLE QList<EnumList::Enum> list() const { return { Alpha, Beta, Gamma }; }
+};
+
+
+class ValueTypeWithEnum1
+{
+    Q_GADGET
+    Q_PROPERTY(ValueTypeWithEnum1::Quality quality READ quality WRITE setQuality)
+public:
+    enum Quality
+    {
+        VeryLowQuality,
+        LowQuality,
+        NormalQuality,
+        HighQuality,
+        VeryHighQuality
+    };
+    Q_ENUM(Quality)
+
+    Quality quality() const { return m_quality; }
+    void setQuality(Quality quality) { m_quality = quality; }
+
+private:
+    Quality m_quality = HighQuality;
+};
+
+class ObjectTypeHoldingValueType1 : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(ValueTypeWithEnum1 vv READ vv WRITE setVv NOTIFY vvChanged)
+    Q_PROPERTY(ValueTypeWithEnum1::Quality q READ q CONSTANT)
+
+public:
+    ValueTypeWithEnum1 vv() const
+    {
+        return m_vv;
+    }
+
+    void setVv(ValueTypeWithEnum1 vv)
+    {
+        if (m_vv.quality() == vv.quality())
+            return;
+
+        m_vv = vv;
+        emit vvChanged(m_vv);
+    }
+
+    ValueTypeWithEnum1::Quality q() const { return m_vv.quality(); }
+
+signals:
+    void vvChanged(ValueTypeWithEnum1 vv);
+
+private:
+    ValueTypeWithEnum1 m_vv;
+};
+
+struct ValueTypeWithEnumForeign1
+{
+    Q_GADGET
+    QML_FOREIGN(ValueTypeWithEnum1)
+    QML_NAMED_ELEMENT(valueTypeWithEnum1)
+};
+
+namespace ValueTypeWithEnumForeignNamespace1
+{
+    Q_NAMESPACE
+    QML_FOREIGN_NAMESPACE(ValueTypeWithEnum1)
+    QML_NAMED_ELEMENT(ValueTypeWithEnum1)
+};
+
+struct ObjectTypeHoldingValueTypeForeign1
+{
+    Q_GADGET
+    QML_FOREIGN(ObjectTypeHoldingValueType1)
+    QML_NAMED_ELEMENT(ObjectTypeHoldingValueType1)
+};
+
+class ValueTypeWithEnum2
+{
+    Q_GADGET
+    Q_PROPERTY(ValueTypeWithEnum2::Quality quality READ quality WRITE setQuality)
+public:
+    enum Quality
+    {
+        VeryLowQuality,
+        LowQuality,
+        NormalQuality,
+        HighQuality,
+        VeryHighQuality
+    };
+    Q_ENUM(Quality)
+
+    Quality quality() const { return m_quality; }
+    void setQuality(Quality quality) { m_quality = quality; }
+
+private:
+    Quality m_quality = HighQuality;
+};
+
+class ObjectTypeHoldingValueType2 : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(ValueTypeWithEnum2 vv READ vv WRITE setVv NOTIFY vvChanged)
+    Q_PROPERTY(ValueTypeWithEnum2::Quality q READ q CONSTANT)
+    Q_CLASSINFO("RegisterEnumsFromRelatedTypes", "false")
+
+public:
+    ValueTypeWithEnum2 vv() const
+    {
+        return m_vv;
+    }
+
+    void setVv(ValueTypeWithEnum2 vv)
+    {
+        if (m_vv.quality() == vv.quality())
+            return;
+
+        m_vv = vv;
+        emit vvChanged(m_vv);
+    }
+
+    ValueTypeWithEnum2::Quality q() const { return m_vv.quality(); }
+
+signals:
+    void vvChanged(ValueTypeWithEnum2 vv);
+
+private:
+    ValueTypeWithEnum2 m_vv;
+};
+
+struct ValueTypeWithEnumForeign2
+{
+    Q_GADGET
+    QML_FOREIGN(ValueTypeWithEnum2)
+    QML_NAMED_ELEMENT(valueTypeWithEnum2)
+};
+
+namespace ValueTypeWithEnumForeignNamespace2
+{
+    Q_NAMESPACE
+    QML_FOREIGN_NAMESPACE(ValueTypeWithEnum2)
+    QML_NAMED_ELEMENT(ValueTypeWithEnum2)
+};
+
+struct ObjectTypeHoldingValueTypeForeign2
+{
+    Q_GADGET
+    QML_FOREIGN(ObjectTypeHoldingValueType2)
+    QML_NAMED_ELEMENT(ObjectTypeHoldingValueType2)
+};
+
+struct Large {
+    Q_GADGET
+    QML_VALUE_TYPE(large)
+
+    Q_PROPERTY(uint a MEMBER a)
+    Q_PROPERTY(uint b MEMBER b)
+    Q_PROPERTY(uint c MEMBER c)
+    Q_PROPERTY(uint d MEMBER d)
+    Q_PROPERTY(uint e MEMBER e)
+    Q_PROPERTY(uint f MEMBER f)
+
+public:
+    quint64 a;
+    quint64 b;
+    quint64 c;
+    quint64 d;
+    quint64 e;
+    quint64 f;
+};
+
+inline bool operator==(const Large &a, const Large &b)
+{
+    return a.a == b.a && a.b == b.b && a.c == b.c && a.d == b.d && a.e == b.e && a.f == b.f;
+}
+
+inline bool operator!=(const Large &a, const Large &b) { return !(a == b); }
+
+class Foo: public QObject {
+
+    Q_OBJECT
+    Q_PROPERTY(QVariantList fooProperty READ getList WRITE setList)
+    Q_PROPERTY(Large a MEMBER a BINDABLE aBindable)
+    Q_PROPERTY(Large b MEMBER b BINDABLE bBindable)
+    QML_ELEMENT
+
+public:
+    QVariantList getList() const { return mFooProperty;}
+    void setList(QVariantList list) { mFooProperty = list;}
+
+    QBindable<Large> aBindable() { return QBindable<Large>(&a); }
+    QBindable<Large> bBindable() { return QBindable<Large>(&b); }
+
+private:
+    QProperty<Large> a;
+    QProperty<Large> b;
+    QVariantList mFooProperty;
+};
+
+struct BaseValueType
+{
+    Q_GADGET
+    Q_PROPERTY(int content READ content WRITE setContent)
+    QML_VALUE_TYPE(base)
+
+public:
+    Q_INVOKABLE void increment() { ++m_content; }
+    Q_INVOKABLE QString report() const { return QString::number(m_content); }
+
+    int content() const { return m_content; }
+    void setContent(int content) { m_content = content; }
+
+private:
+    int m_content = 27;
+};
+
+struct DerivedValueType : public BaseValueType
+{
+    Q_GADGET
+    QML_VALUE_TYPE(derived)
+public:
+    DerivedValueType() { increment(); }
+    Q_INVOKABLE int nothing() const { return m_nothing; }
+
+private:
+    int m_nothing = 12;
+};
+
+class ItemAttached : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(QString attachedName READ attachedName WRITE setAttachedName NOTIFY attachedNameChanged)
+    QML_ELEMENT
+    QML_ATTACHED(ItemAttached)
+public:
+    ItemAttached(QObject *parent = nullptr) : QObject(parent) {}
+
+    QString attachedName() const { return m_name; }
+    void setAttachedName(const QString &name)
+    {
+        if (name != m_name) {
+            m_name = name;
+            emit attachedNameChanged();
+        }
+    }
+
+    static ItemAttached *qmlAttachedProperties(QObject *object)
+    {
+        if (object->objectName() != QLatin1String("foo")) {
+            qWarning("Only foo can have ItemAttached!");
+            return nullptr;
+        }
+
+        return new ItemAttached(object);
+    }
+
+signals:
+    void attachedNameChanged();
+
+private:
+    QString m_name;
+};
+
+class BindableOnly : public QObject
+{
+    Q_OBJECT
+    Q_PROPERTY(QByteArray data READ data WRITE setData BINDABLE dataBindable FINAL)
+    QML_ELEMENT
+public:
+    BindableOnly(QObject *parent = nullptr)
+        : QObject(parent)
+    {}
+
+    QBindable<QByteArray> dataBindable() { return QBindable<QByteArray>(&m_data); }
+
+    QByteArray data() const { return m_data.value(); }
+    void setData(const QByteArray &newData) { m_data.setValue(newData); }
+
+private:
+    QProperty<QByteArray> m_data;
+};
+
 void registerTypes();
 
 #endif // TESTTYPES_H
